// Copyright (C) 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// * The software is provided "as is" without any warranty, neither express
//   nor implied.
// * In no event will the author be held liable for any damage(s) arising
//   from the use of the software.
// * Redistribution of the software is allowed only in unmodified form.
// * Permission is granted to use the software solely for personal private
//   and non-commercial purposes.
// * An individuals use of the software in his or her capacity or function
//   as an agent, (independent) contractor, employee, member or officer of
//   a business, corporation or organization (commercial or non-commercial)
//   does not qualify as personal private and non-commercial purpose.
// * Without written approval from the author the software must not be used
//   for a business, for commercial, corporate, governmental, military or
//   organizational purposes of any kind, or in a commercial, corporate,
//   governmental, military or organizational environment of any kind.

// Minimalist Win32 Runtime Library for Microsoft C Compiler 20xy

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(linker, "/DEFAULTLIB:KERNEL32.LIB")
#pragma comment(linker, "/DEFAULTLIB:USER32.LIB")
#pragma comment(user, __TIMESTAMP__)

#ifndef _VC_NODEFAULTLIB
#pragma message("Use '/Zl' compiler switch!")
#endif

#define STRICT
#define WIN32_LEAN_AND_MEAN

#define WINVER		0x0500
#define _WIN32_WINNT	0x0500

#include <windows.h>
#include "nomsvcrt.h"

extern	const	IMAGE_DOS_HEADER	__ImageBase;

const	LPCWSTR	szReason[5] = {L"DLL_PROCESS_DETACH",
		               L"DLL_PROCESS_ATTACH",
		               L"DLL_THREAD_ATTACH",
		               L"DLL_THREAD_DETACH",
		               L"DLL_PROCESS_VERIFIER"};

#ifdef _DLL
// https://msdn.microsoft.com/en-us/library/ms682583.aspx
// https://msdn.microsoft.com/en-us/library/ms682596.aspx

BOOL	WINAPI	DllMain(HANDLE hModule,
		        DWORD  dwReason,
		        LPVOID lpReserved)
{
	// https://msdn.microsoft.com/en-us/library/ms686708.aspx
	// https://msdn.microsoft.com/en-us/library/hh285210.aspx

	NT_TIB	*tib = (NT_TIB *) NtCurrentTeb();

	// https://msdn.microsoft.com/en-us/library/ms683197.aspx

	WCHAR	szModule[MAX_PATH];
	DWORD	dwModule = GetModuleFileNameW(hModule, szModule, sizeof(szModule) / sizeof(*szModule));

	// https://msdn.microsoft.com/en-us/library/ms683199.aspx

	HMODULE	hProcess = GetModuleHandleW((LPCWSTR) NULL);
	WCHAR	szProcess[MAX_PATH];
	DWORD	dwProcess = GetModuleFileNameW(hProcess, szProcess, sizeof(szProcess) / sizeof(*szProcess));

	WCHAR	szBuffer[1024];
	DWORD	dwBuffer = 0UL;

	DWORD	dwButton = 0UL;

	if (dwModule < sizeof(szModule) / sizeof(*szModule))
		szModule[dwModule] = L'\0';

	if (dwProcess < sizeof(szProcess) / sizeof(*szProcess))
		szProcess[dwProcess] = L'\0';

	// https://msdn.microsoft.com/en-us/library/ms683224.aspx

	dwProcess = GetProcessVersion(0UL);

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	dwBuffer = wsprintfW(szBuffer, L"Called address = 0x%p\n"
	                               L"Return address = 0x%p\n"
	                               L"Address of return address = 0x%p\n"
	                               L"Stack base = 0x%p\n"
	                               L"Stack pointer = 0x%p\n"
	                               L"Stack limit = 0x%p\n"
	                               L"\n"
	                               L"Arguments:\n"
	                               L"\tHandle = 0x%p\n"
	                               L"\tReason = %ls\n"
	                               L"\tContext = 0x%p\n"
	                               L"\n"
	                               L"DLL image base = 0x%p\n"
	                               L"DLL filename = %ls\n"
	                               L"\n"
	                               L"Process image base = 0x%p\n"
	                               L"Process filename = %ls\n"
	                               L"Process version = %hu.%hu\n"
	                               L"Process id = %lu\n"
	                               L"\n"
	                               L"Thread information block = 0x%p\n"
	                               L"Thread id = %lu\n",
	                     DllMain,
	                     _ReturnAddress(),
	                     _AddressOfReturnAddress(),
	                     tib->StackBase,
	                     _alloca(sizeof(LPVOID)),
	                     tib->StackLimit,
	                     hModule,
	                     szReason[dwReason],
	                     lpReserved,
	                     &__ImageBase,
	                     szModule,
	                     hProcess,
	                     szProcess,
	                     HIWORD(dwProcess), LOWORD(dwProcess),
	                     GetCurrentProcessId(),
	                     tib->Self,
	                     GetCurrentThreadId());

	if (dwBuffer == 0UL)
		return FALSE;

	szBuffer[dwBuffer] = L'\0';

	// https://msdn.microsoft.com/en-us/library/ms645507.aspx

	dwButton = MessageBoxExW(HWND_DESKTOP,
	                         szBuffer,
	                         L"DllMain() function",
	                         dwReason == DLL_PROCESS_ATTACH ? MB_OKCANCEL : MB_OK,
	                         MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));

	return dwButton == IDOK ? TRUE : FALSE;
}
#else // _DLL
#ifdef CONSOLE
#ifdef UNICODE
// https://msdn.microsoft.com/en-us/library/6wd819wh.aspx

int	__cdecl	wmain(int argc, wchar_t *argv[], wchar_t *envp[])
{
	PrintFormatW(STDOUT, L"argc:\t\t%lu\n", argc);

	argc = 0;

	while (*argv != NULL)
		PrintFormatW(STDOUT, L"argv[%lu]:\t0x%p\t%ls\n", argc++, *argv++, *argv);

	PrintCharW(STDOUT, L'\n');

	argc = 0;

	while (*envp != NULL)
		PrintFormatW(STDOUT, L"envp[%lu]:\t0x%p\t%ls\n", argc++, *envp++, *envp);

	return argc;
}
#else // ANSI
// https://msdn.microsoft.com/en-us/library/3ze4ytsc.aspx
// https://msdn.microsoft.com/en-us/library/6wd819wh.aspx

int	__cdecl	main(int argc, char *argv[], char *envp[])
{
	PrintFormatA(STDOUT, "argc:\t\t%lu\n", argc);

	argc = 0;

	while (*argv != NULL)
		PrintFormatA(STDOUT, "argv[%lu]:\t0x%p\t%hs\n", argc++, *argv++, *argv);

	PrintCharA(STDOUT, '\n');

	argc = 0;

	while (*envp != NULL)
		PrintFormatA(STDOUT, "envp[%lu]:\t0x%p\t%hs\n", argc++, *envp++, *envp);

	return argc;
}
#endif // UNICODE
#else // WINDOWS
#ifdef UNICODE
// https://msdn.microsoft.com/en-us/library/ms633559.aspx
// https://msdn.microsoft.com/en-us/library/ff381406.aspx

INT	WINAPI	wWinMain(HINSTANCE hInstance,
		         HINSTANCE hPrevInstance,
	                 LPWSTR    lpCmdLine,
	                 INT       nCmdShow)
{
	// https://msdn.microsoft.com/en-us/library/ms686708.aspx
	// https://msdn.microsoft.com/en-us/library/hh285210.aspx

	NT_TIB	*tib = (NT_TIB *) NtCurrentTeb();

	// https://msdn.microsoft.com/en-us/library/ms683197.aspx

	WCHAR	szProcess[MAX_PATH];
	DWORD	dwProcess = GetModuleFileNameW((HMODULE) &__ImageBase, szProcess, sizeof(szProcess) / sizeof(*szProcess));

	WCHAR	szBuffer[1024];
	DWORD	dwBuffer = 0UL;

	if (dwProcess < sizeof(szProcess) / sizeof(*szProcess))
		szProcess[dwProcess] = L'\0';

	// https://msdn.microsoft.com/en-us/library/ms683224.aspx

	dwProcess = GetProcessVersion(0UL);

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	dwBuffer = wsprintfW(szBuffer, L"Called address = 0x%p\n"
	                               L"Return address = 0x%p\n"
	                               L"Address of return address = 0x%p\n"
	                               L"Stack base = 0x%p\n"
	                               L"Stack pointer = 0x%p\n"
	                               L"Stack limit = 0x%p\n"
	                               L"\n"
	                               L"Arguments:\n"
	                               L"\tInstance = 0x%p\n"
	                               L"\tPrevInstance = 0x%p\n"
	                               L"\tCmdLine = %ls\n"
	                               L"\tCmdShow = 0x%p\n"
	                               L"\n"
	                               L"Process image base = 0x%p\n"
	                               L"Process filename = %ls\n"
	                               L"Process version = %hu.%hu\n"
	                               L"Process id = %lu\n"
	                               L"\n"
	                               L"Thread information block = 0x%p\n"
	                               L"Thread id = %lu\n",
	                     wWinMain,
	                     _ReturnAddress(),
	                     _AddressOfReturnAddress(),
	                     tib->StackBase,
	                     _alloca(sizeof(LPVOID)),
	                     tib->StackLimit,
	                     hInstance,
	                     hPrevInstance,
	                     lpCmdLine,
	                     nCmdShow,
	                     &__ImageBase,
	                     szProcess,
	                     HIWORD(dwProcess), LOWORD(dwProcess),
	                     GetCurrentProcessId(),
	                     tib->Self,
	                     GetCurrentThreadId());

	if (dwBuffer == 0UL)
		return -1;

	szBuffer[dwBuffer] = L'\0';

	// https://msdn.microsoft.com/en-us/library/ms645507.aspx

	return MessageBoxExW(HWND_DESKTOP,
	                     szBuffer,
	                     L"wWinMain() function",
	                     MB_OK,
	                     MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));
}
#else // ANSI
// https://msdn.microsoft.com/en-us/library/ms633559.aspx
// https://msdn.microsoft.com/en-us/library/ff381406.aspx

INT	WINAPI	WinMain(HINSTANCE hInstance,
		        HINSTANCE hPrevInstance,
	                LPSTR     lpCmdLine,
	                INT       nCmdShow)
{

	// https://msdn.microsoft.com/en-us/library/ms645507.aspx

	return MessageBoxExA(HWND_DESKTOP,
	                     lpCmdLine,
	                     "WinMain() function",
	                     MB_OK,
	                     MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));
}
#endif // UNICODE
#endif // CONSOLE
#endif // _DLL
